/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy;

import static java.util.Objects.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.concurrent.atomic.AtomicReference;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.Attachment;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessage.BayernIdMessageBuilder;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessageMapper;
import io.grpc.stub.StreamObserver;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@RequiredArgsConstructor
public class UploadStreamObserver implements StreamObserver<GrpcSendBayernIdMessageRequest> {

	private final StreamObserver<GrpcSendBayernIdMessageResponse> responseObserver;
	private final BayernIdProxyService proxyService;
	private final BayernIdMessageMapper messageMapper;

	private final AtomicReference<BayernIdMessageBuilder> messageBuilder = new AtomicReference<>();
	private OutputStream attachmentWriter;

	@Override
	public synchronized void onNext(GrpcSendBayernIdMessageRequest sendRequest) {
		if (sendRequest.hasMessageMetadata()) {
			readMessageMetadata(sendRequest.getMessageMetadata());
			return;
		}
		readAttachments(sendRequest.getAttachments());
	}

	void readMessageMetadata(GrpcBayernIdMessageMetadata sendRequest) {
		messageBuilder.set(messageMapper.toBayernIdMessage(sendRequest));
	}

	void readAttachments(GrpcAttachments attachments) {
		if (attachments.hasAttachmentMetadata()) {
			addDataContainer(attachments.getAttachmentMetadata());
			return;
		}
		writeAttachment(attachments.getContent());
	}

	void addDataContainer(GrpcAttachmentMetadata attachmentMetadata) {
		messageBuilder.get().attachment(buildAttachment(attachmentMetadata));
	}

	Attachment buildAttachment(GrpcAttachmentMetadata attachmentMetadata) {
		InputStream temporallyFile;
		try {
			temporallyFile = createTemporallyFile();
		} catch (IOException e) {
			throw new TechnicalException("Can not create temporary file for attachment " + attachmentMetadata.getFileName(), e);
		}
		return Attachment.builder()
				.name(attachmentMetadata.getFileName())
				.type(attachmentMetadata.getFileType())
				.content(temporallyFile)
				.build();
	}

	InputStream createTemporallyFile() throws IOException {
		closeTemporallyWriter();
		var tempFilePath = TempFileUtils.createTmpFile();
		attachmentWriter = newOutputStream(tempFilePath);
		return newInputStream(tempFilePath);
	}

	OutputStream newOutputStream(Path filePath) throws IOException {
		return Files.newOutputStream(filePath);
	}

	InputStream newInputStream(Path filePath) throws IOException {
		return new DeleteOnCloseInputStream(filePath.toFile());
	}

	void closeTemporallyWriter() {
		try {
			if (nonNull(attachmentWriter)) {
				attachmentWriter.close();
			}
		} catch (IOException e) {
			throw new TechnicalException("Can not close attachment file", e);
		}
	}

	void writeAttachment(ByteString content) {
		try {
			attachmentWriter.write(content.toByteArray());
			attachmentWriter.flush();
		} catch (IOException e) {
			throw new TechnicalException("Can not save attachment content", e);
		}
	}

	@Override
	public void onError(Throwable t) {
		closeTemporallyWriter();
		throw new TechnicalException("Error while sending postfach message", t);
	}

	@Override
	public void onCompleted() {
		closeTemporallyWriter();
		try {
			var bayernIdResponse = proxyService.sendPostfachNachricht(messageBuilder.get().build());
			responseObserver.onNext(messageMapper.fromBayernIdResponse(bayernIdResponse));
			responseObserver.onCompleted();
		} catch (Exception e) {
			throw new TechnicalException("Error while sending postfach message", e);
		}
	}
}
