/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy;

import static org.assertj.core.api.Assertions.*;

import java.io.InputStream;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import org.apache.commons.lang3.tuple.Pair;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.ImportAutoConfiguration;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.Resource;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import com.github.tomakehurst.wiremock.junit5.WireMockRuntimeInfo;
import com.github.tomakehurst.wiremock.junit5.WireMockTest;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.akdb.BayernIdRemoteService;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.mock.GrpcBayernIdProxyTestClient;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.mock.MockBayernIdInitializer;
import io.grpc.StatusRuntimeException;
import lombok.SneakyThrows;
import net.devh.boot.grpc.client.autoconfigure.GrpcClientAutoConfiguration;
import net.devh.boot.grpc.client.autoconfigure.GrpcClientHealthAutoConfiguration;
import net.devh.boot.grpc.client.autoconfigure.GrpcClientMetricAutoConfiguration;
import net.devh.boot.grpc.client.autoconfigure.GrpcClientSecurityAutoConfiguration;
import net.devh.boot.grpc.client.autoconfigure.GrpcClientTraceAutoConfiguration;
import net.devh.boot.grpc.client.autoconfigure.GrpcDiscoveryClientAutoConfiguration;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(classes = BayernIdProxyApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.inProcess.address=in-process:test",
		"ozgcloud.bayernid.server: http://bayern-id",
		"ozgcloud.bayernid.useSsl: false"
})
@SpringJUnitConfig(classes = BayernIdProxyGrpcServiceITCase.TestConfig.class)
@DirtiesContext
@WireMockTest
@ITCase
class BayernIdProxyGrpcServiceITCase {

	@Autowired
	private BayernIdProxyGrpcService service;
	@Autowired
	private BayernIdRemoteService bayernIdRemoteService;

	@GrpcClient("inProcess")
	private BayernIdProxyServiceGrpc.BayernIdProxyServiceStub serviceStub;

	private MockBayernIdInitializer bayernIdInitializer;

	@Value("classpath:test.pdf")
	private Resource pdfResource;
	@Value("classpath:test.txt")
	private Resource txtResource;

	private String pdfName = "test.pdf";
	private String txtName = "test.txt";

	private GrpcBayernIdProxyTestClient grpcTestClient;

	@BeforeEach
	void setup(WireMockRuntimeInfo wmRuntimeInfo) {
		bayernIdRemoteService.setDefaultUri("http://localhost:" + wmRuntimeInfo.getHttpPort());
		bayernIdInitializer = new MockBayernIdInitializer(wmRuntimeInfo);
		grpcTestClient = new GrpcBayernIdProxyTestClient(serviceStub);
	}

	@Test
	@SneakyThrows
	void shouldSendMessage() {
		bayernIdInitializer.createResponseOk();

		var bayernIdResponse = grpcTestClient.sendMessage(GrpcBayernIdMessageMetadataTestFactory.create(), createAttachmentIterator());

		assertThat(bayernIdResponse.isSuccess()).isTrue();
		assertThat(bayernIdResponse.getStatus()).isEqualTo("0");
	}

	@SneakyThrows
	@Test
	void shouldReceiveErrorResponse() {
		bayernIdInitializer.createResponseError();

		var bayernIdResponse = grpcTestClient.sendMessage(GrpcBayernIdMessageMetadataTestFactory.create(), createAttachmentIterator());

		assertThat(bayernIdResponse.isSuccess()).isFalse();
		assertThat(bayernIdResponse.getStatus()).isEqualTo("31");
		assertThat(bayernIdResponse.getMessage()).isEqualTo("Unzulässiger Nachrichteninhalt");
	}

	@Test
	void shouldPassException() {
		bayernIdInitializer.createResponseWithException();

		try {
			grpcTestClient.sendMessage(GrpcBayernIdMessageMetadataTestFactory.create(), createAttachmentIterator());
		} catch (ExecutionException e) {
			assertException(e.getMessage(), e.getCause());
		}
	}

	@Test
	void shouldPass500Response() {
		bayernIdInitializer.create500Response();

		try {
			grpcTestClient.sendMessage(GrpcBayernIdMessageMetadataTestFactory.create(), createAttachmentIterator());
		} catch (ExecutionException e) {
			assertException(e.getMessage(), e.getCause());
		}
	}

	private void assertException(String exceptionMessage, Throwable cause) {
		assertThat(cause).isInstanceOf(StatusRuntimeException.class);
		assertThat(exceptionMessage).contains("(ExceptionId:");
	}


	@SneakyThrows
	Set<Pair<GrpcAttachmentMetadata, InputStream>> createAttachmentIterator() {
		return Set.of(
				Pair.of(GrpcAttachmentMetadata.newBuilder().setFileName(txtName).setFileType("text/plain").build(), txtResource.getInputStream()),
				Pair.of(GrpcAttachmentMetadata.newBuilder().setFileName(pdfName).setFileType("application/pdf").build(), pdfResource.getInputStream())
		);
	}

	@Configuration
	@ImportAutoConfiguration({
			GrpcClientAutoConfiguration.class,
			GrpcClientHealthAutoConfiguration.class,
			GrpcClientMetricAutoConfiguration.class,
			GrpcClientSecurityAutoConfiguration.class,
			GrpcClientTraceAutoConfiguration.class,
			GrpcDiscoveryClientAutoConfiguration.class
	})
	static class TestConfig {
	}
}