/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy.akdb;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import jakarta.xml.bind.JAXBElement;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.ws.client.core.WebServiceTemplate;

import akdb.bsp.postkorb.komm.webservice.ObjectFactory;
import akdb.bsp.postkorb.komm.webservice.SendBspNachrichtNative;
import akdb.bsp.postkorb.komm.webservice.SendBspNachrichtNativeOutput;
import de.akdb.egov.bsp.nachrichten.BspNachricht;
import de.akdb.egov.bsp.nachrichten.BspQuittung;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessage;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessageMapper;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdResponse;

class BayernIdRemoteServiceTest {

	@Spy
	@InjectMocks
	private BayernIdRemoteService service;

	@Mock
	private BayernIdMessageMapper messageMapper;
	@Mock
	private ObjectFactory objectFactory;

	@Nested
	class TestSendPostfachNachricht {

		@Mock
		private SendBspNachrichtNative sendNachricht;
		@Mock
		private SendBspNachrichtNativeOutput sendBspNachrichtResponse;
		@Mock
		private BspQuittung bspQuittung;
		@Mock
		private JAXBElement<SendBspNachrichtNativeOutput> jaxbSendBspNachrichtNativeOutput;
		@Mock
		private BayernIdResponse response;
		@Mock
		private BayernIdMessage message;

		@BeforeEach
		void setup(){
			doReturn(bspQuittung).when(service).createBspQuitung(any());
			doReturn(sendNachricht).when(service).createSendNachricht(any());
		}

		@Test
		void shouldCallCreateSendNachricht() {
			doReturn(jaxbSendBspNachrichtNativeOutput).when(service).send(any());

			sendMessage();

			verify(service).createSendNachricht(message);
		}

		@Test
		void shouldCallSendBspNachricht() {
			doReturn(jaxbSendBspNachrichtNativeOutput).when(service).send(any());

			sendMessage();

			verify(service).sendBspNachricht(sendNachricht);
		}

		@Test
		void shouldCallCreateBspQuitung() {
			doReturn(sendBspNachrichtResponse).when(service).sendBspNachricht(any());

			sendMessage();

			verify(service).createBspQuitung(sendBspNachrichtResponse);
		}

		@Test
		void shouldCallMessageMapper() {
			doReturn(sendBspNachrichtResponse).when(service).sendBspNachricht(any());

			sendMessage();

			verify(messageMapper).toBayernIdResponse(bspQuittung);
		}

		@Test
		void shouldReturnResponse() {
			doReturn(response).when(messageMapper).toBayernIdResponse(bspQuittung);
			doReturn(sendBspNachrichtResponse).when(service).sendBspNachricht(any());

			var result = sendMessage();

			assertThat(result).isEqualTo(response);
		}

		private BayernIdResponse sendMessage() {
			return service.sendPostfachNachricht(message);
		}

	}

	@Nested
	class TestCreateSendNachricht {

		private static final String RESULT_STRING = "result";

		@Mock
		private BayernIdMessage message;
		@Mock
		private BspNachricht bspNachricht;

		@BeforeEach
		void setup() {
			doReturn(RESULT_STRING).when(service).buildBspNachrichtXml(any());
		}

		@Test
		void shouldCallMessageMapper() {
			service.createSendNachricht(message);

			verify(messageMapper).toBspNachricht(message);
		}

		@Test
		void shouldCallBuildBspNachrichtXml() {
			when(messageMapper.toBspNachricht(any())).thenReturn(bspNachricht);

			service.createSendNachricht(message);

			verify(service).buildBspNachrichtXml(bspNachricht);
		}

		@Test
		void shouldSetBspNachricht() {
			var result = service.createSendNachricht(message);

			assertThat(result.getBspNachricht()).isEqualTo(RESULT_STRING);
		}
	}

	@Nested
	class TestSendBspNachricht {

		@Mock
		private SendBspNachrichtNative sendBspNachrichtRequest;
		@Mock
		private JAXBElement<SendBspNachrichtNative> jaxbSendBspNachrichtNative;
		@Mock
		private JAXBElement<SendBspNachrichtNativeOutput> jaxbSendBspNachrichtNativeOutput;
		@Mock
		private SendBspNachrichtNativeOutput response;

		@BeforeEach
		void setup() {
			doReturn(jaxbSendBspNachrichtNativeOutput).when(service).send(any());
			when(jaxbSendBspNachrichtNativeOutput.getValue()).thenReturn(response);
		}

		@Test
		void shouldCallMarshaller() {
			service.sendBspNachricht(sendBspNachrichtRequest);

			verify(objectFactory).createSendBspNachrichtNative(sendBspNachrichtRequest);
		}

		@Test
		void shouldReturnValue() {
			var sendBspNachrichtNativeOutput = service.sendBspNachricht(sendBspNachrichtRequest);

			assertThat(sendBspNachrichtNativeOutput).isEqualTo(response);
		}

		@Test
		void shouldCallSend() {
			when(objectFactory.createSendBspNachrichtNative(any())).thenReturn(jaxbSendBspNachrichtNative);

			service.sendBspNachricht(sendBspNachrichtRequest);

			verify(service).send(jaxbSendBspNachrichtNative);
		}

	}

	@Nested
	class TestSend {

		@Mock
		private JAXBElement<SendBspNachrichtNative> jaxbSendBspNachrichtNative;
		@Mock
		private WebServiceTemplate webServiceTemplate;

		@Test
		void shouldCallMarshalSendAndReceive() {
			doReturn(webServiceTemplate).when(service).getWebServiceTemplate();

			service.send(jaxbSendBspNachrichtNative);

			verify(webServiceTemplate).marshalSendAndReceive(jaxbSendBspNachrichtNative);
		}

	}
}