/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy.errorhandling;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.grpc.GrpcUtil;
import io.grpc.Metadata;
import io.grpc.Status;
import io.grpc.Status.Code;
import io.grpc.StatusException;

class ExceptionHandlerTest {

	@Spy
	@InjectMocks
	private ExceptionHandler exceptionHandler;

	@Nested
	class TestHandleTechnicalException {

		private static final TechnicalException EXCEPTION = new TechnicalException("Test");

		@Mock
		private Metadata metadata;
		@Mock
		private StatusException statusException;

		@Test
		void shouldCallBuildInternalStatus() {
			exceptionHandler.handleTechnicalException(EXCEPTION);

			verify(exceptionHandler).buildInternalStatus(EXCEPTION);
		}

		@Test
		void shouldCallBuildMetadata() {
			exceptionHandler.handleTechnicalException(EXCEPTION);

			verify(exceptionHandler).buildMetadata(EXCEPTION.getExceptionId());
		}

		@Test
		void shouldCallCreateStatusException() {
			var status = Status.INTERNAL;
			doReturn(status).when(exceptionHandler).buildInternalStatus(any());
			doReturn(metadata).when(exceptionHandler).buildMetadata(any());

			exceptionHandler.handleTechnicalException(EXCEPTION);

			verify(exceptionHandler).createStatusException(status, metadata);
		}

		@Test
		void shouldReturnStatusException() {
			doReturn(statusException).when(exceptionHandler).createStatusException(any(), any());

			var result = exceptionHandler.handleTechnicalException(EXCEPTION);

			assertThat(result).isEqualTo(statusException);
		}
	}

	@Nested
	class TestHandleRuntimeException {

		private static final RuntimeException EXCEPTION = new RuntimeException("Test");

		@Mock
		private Metadata metadata;
		@Mock
		private StatusException statusException;

		@Test
		void shouldCallBuildInternalStatus() {
			exceptionHandler.handleRuntimeException(EXCEPTION);

			verify(exceptionHandler).buildInternalStatus(EXCEPTION);
		}

		@Test
		void shouldCallCreateExceptionId() {
			exceptionHandler.handleRuntimeException(EXCEPTION);

			verify(exceptionHandler).createExceptionId();
		}

		@Test
		void shouldCallBuildMetadata() {
			var exceptionId = "42";
			doReturn(exceptionId).when(exceptionHandler).createExceptionId();

			exceptionHandler.handleRuntimeException(EXCEPTION);

			verify(exceptionHandler).buildMetadata(exceptionId);
		}

		@Test
		void shouldCallCreateStatusException() {
			var status = Status.INTERNAL;
			doReturn(status).when(exceptionHandler).buildInternalStatus(any());
			doReturn(metadata).when(exceptionHandler).buildMetadata(any());

			exceptionHandler.handleRuntimeException(EXCEPTION);

			verify(exceptionHandler).createStatusException(status, metadata);
		}

		@Test
		void shouldReturnStatusException() {
			doReturn(statusException).when(exceptionHandler).createStatusException(any(), any());

			var result = exceptionHandler.handleRuntimeException(EXCEPTION);

			assertThat(result).isEqualTo(statusException);
		}
	}

	@Nested
	class TestBuildInternalStatus {

		private static final Exception CAUSE = new Exception("Cause");
		private static final RuntimeException EXCEPTION = new RuntimeException("Test", CAUSE);

		@Test
		void shouldSetInternalStatusCode() {
			var result = exceptionHandler.buildInternalStatus(EXCEPTION);

			assertThat(result.getCode()).isEqualTo(Code.INTERNAL);
		}

		@Test
		void shouldSetDescription() {
			var result = exceptionHandler.buildInternalStatus(EXCEPTION);

			assertThat(result.getDescription()).isEqualTo(EXCEPTION.getMessage());
		}

		@Test
		void shouldSetCause() {
			var result = exceptionHandler.buildInternalStatus(EXCEPTION);

			assertThat(result.getCause()).isEqualTo(CAUSE);
		}
	}

	@Nested
	class TestBuildMetadata {

		private static final String EXCEPTION_ID = "42";
		@Test
		void shouldSetExceptionIdKey() {
			var result = exceptionHandler.buildMetadata(EXCEPTION_ID);

			assertThat(result.keys()).containsOnly(ExceptionHandler.KEY_EXCEPTION_ID.toLowerCase());
		}

		@Test
		void shouldSetExceptionIdValue() {
			var result = exceptionHandler.buildMetadata(EXCEPTION_ID);

			assertThat(result.get(GrpcUtil.keyOfString(ExceptionHandler.KEY_EXCEPTION_ID))).isEqualTo(EXCEPTION_ID);
		}

	}
}