/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy.message;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.util.List;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.Mockito;
import org.testcontainers.shaded.org.apache.commons.io.IOUtils;

import de.akdb.egov.bsp.nachrichten.BspNachricht;
import de.akdb.egov.bsp.nachrichten.BspQuittung;
import de.akdb.egov.bsp.nachrichten.DataContainerType;
import de.akdb.egov.bsp.nachrichten.FreiTextType;
import de.akdb.egov.bsp.nachrichten.NachrichtenInhaltType;
import de.akdb.egov.bsp.nachrichten.NachrichtenKopfType;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.BayernIdResponseTestFactory;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcAbsenderTestFactory;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcBayernIdMessageMetadata;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcBayernIdMessageMetadataTestFactory;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcEmpfaengerTestFactory;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageResponse;
import lombok.SneakyThrows;

class BayernIdMessageMapperTest {

	private BayernIdMessageMapper mapper = Mappers.getMapper(BayernIdMessageMapper.class);

	@Nested
	class ToBayernIdMessage {

		private static final GrpcBayernIdMessageMetadata MESSAGE_METADATA = GrpcBayernIdMessageMetadataTestFactory.create();

		@Test
		void shouldSetMessageId() {
			var result = mapToBayernIdMessage();

			assertThat(result.getMessageId()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.MESSAGE_ID);
		}

		@Test
		void shouldSetCreatedAt() {
			var result = mapToBayernIdMessage();

			assertThat(result.getCreatedAt()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.CREATED_AT_STRING);
		}

		@Test
		void shouldSetSubject() {
			var result = mapToBayernIdMessage();

			assertThat(result.getSubject()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.SUBJECT);
		}

		@Test
		void shoulSetStorkQaaLevel() {
			var result = mapToBayernIdMessage();

			assertThat(result.getStorkQaaLevel()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.STORK_QAA_LEVEL);
		}

		@Test
		void shouldSetVorgangId() {
			var result = mapToBayernIdMessage();

			assertThat(result.getVorgangId()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetText() {
			var result = mapToBayernIdMessage();

			assertThat(result.getText()).isEqualTo(GrpcBayernIdMessageMetadataTestFactory.MESSAGE_TEXT);
		}

		@Nested
		class TestSetAbsender {

			@Test
			void shouldSetName() {
				var result = mapToAbsender();

				assertThat(result.getName()).isEqualTo(GrpcAbsenderTestFactory.NAME);
			}

			@Test
			void shouldSetAnschrift() {
				var result = mapToAbsender();

				assertThat(result.getAnschrift()).isEqualTo(GrpcAbsenderTestFactory.ANSCHRIFT);
			}

			@Test
			void shouldSetEmail() {
				var result = mapToAbsender();

				assertThat(result.getEmail()).isEqualTo(GrpcAbsenderTestFactory.EMAIL);
			}

			@Test
			void shouldSetTelefon() {
				var result = mapToAbsender();

				assertThat(result.getTelefon()).isEqualTo(GrpcAbsenderTestFactory.TELEFON);
			}

			@Test
			void shouldSetHyperlink() {
				var result = mapToAbsender();

				assertThat(result.getHyperlink()).isEqualTo(GrpcAbsenderTestFactory.HYPERLINK);
			}

			@Test
			void shouldSetMandant() {
				var result = mapToAbsender();

				assertThat(result.getMandant()).isEqualTo(GrpcAbsenderTestFactory.MANDANT);
			}

			@Test
			void shouldSetDienst() {
				var result = mapToAbsender();

				assertThat(result.getDienst()).isEqualTo(GrpcAbsenderTestFactory.DIENST);
			}

			@Test
			void shouldSetGemeindeschluessel() {
				var result = mapToAbsender();

				assertThat(result.getGemeindeschluessel()).isEqualTo(GrpcAbsenderTestFactory.GEMEINDE_SCHLUESSEL);
			}

			private Absender mapToAbsender() {
				var bayernIdMessage = mapToBayernIdMessage();
				assertThat(bayernIdMessage.getAbsender()).isNotNull();
				return bayernIdMessage.getAbsender();
			}
		}

		@Nested
		class TestSetEmpfaenger {

			@Test
			void shouldSetPostfachId() {
				var result = mapToEmpfänger();

				assertThat(result.getPostkorbId()).isEqualTo(GrpcEmpfaengerTestFactory.POSTFACH_ID);
			}

			@Test
			void shouldSetName() {
				var result = mapToEmpfänger();

				assertThat(result.getName()).isEqualTo(GrpcEmpfaengerTestFactory.NAME);
			}

			@Test
			void shouldSetAnschrift() {
				var result = mapToEmpfänger();

				assertThat(result.getAnschrift()).isEqualTo(GrpcEmpfaengerTestFactory.ANSCHRIFT);
			}

			private Empfaenger mapToEmpfänger() {
				var bayernIdMessage = mapToBayernIdMessage();
				assertThat(bayernIdMessage.getEmpfaenger()).isNotNull();
				return bayernIdMessage.getEmpfaenger();

			}
		}

		private BayernIdMessage mapToBayernIdMessage() {
			return mapper.toBayernIdMessage(MESSAGE_METADATA).build();
		}
	}

	@Nested
	class TestFromBayernIdResponse {

		private static final BayernIdResponse BAYERN_ID_RESPONSE = BayernIdResponseTestFactory.create();

		@Test
		void shouldSetSuccess() {
			var result = mapToGrpcSendBayernIdMessageResponse();

			assertThat(result.getSuccess()).isTrue();
		}

		@Test
		void shouldSetStatus() {
			var result = mapToGrpcSendBayernIdMessageResponse();

			assertThat(result.getStatus()).isEqualTo(BayernIdResponseTestFactory.STATUS);
		}

		@Test
		void shouldSetMessage() {
			var result = mapToGrpcSendBayernIdMessageResponse();

			assertThat(result.getMessage()).isEqualTo(BayernIdResponseTestFactory.MESSAGE);
		}

		private GrpcSendBayernIdMessageResponse mapToGrpcSendBayernIdMessageResponse() {
			return mapper.fromBayernIdResponse(BAYERN_ID_RESPONSE);
		}
	}

	@Nested
	class TestToBspNachricht {

		@Nested
		class TestSetNachrichtenKopf {

			@Test
			void shouldSetNachrichtId() {
				var result = mapToNachrichtenKopf().getIdentifikationNachricht();

				assertThat(result.getNachrichtenId()).isEqualTo(BayernIdMessageTestFactory.MESSAGE_ID);
			}

			@Test
			void shouldSetErstellungszeitpunkt() {
				var result = mapToNachrichtenKopf().getIdentifikationNachricht();

				assertThat(result.getErstellungszeitpunkt()).hasToString(BayernIdMessageTestFactory.CREATED_AT);
			}

			@Nested
			class TestMapEmpfaenger {

				@Test
				void shouldSetPostkorbId() {
					var result = mapToNachrichtenKopf().getEmpfaenger();

					assertThat(result.getPostkorbId()).isEqualTo(EmpfaengerTestFactory.POSTKORB_ID);
				}

				@Test
				void shouldSetName() {
					var result = mapToNachrichtenKopf().getEmpfaenger();

					assertThat(result.getName()).isEqualTo(EmpfaengerTestFactory.NAME);
				}

				@Test
				void shouldSetAnschrift() {
					var result = mapToNachrichtenKopf().getEmpfaenger();

					assertThat(result.getAnschrift()).isEqualTo(EmpfaengerTestFactory.ANSCHRIFT);
				}
			}

			@Nested
			class TestMapAbsender {

				@Test
				void shouldSetName() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getName()).isEqualTo(AbsenderTestFactory.NAME);
				}

				@Test
				void shouldSetAnschrift() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getAnschrift()).isEqualTo(AbsenderTestFactory.ANSCHRIFT);
				}

				@Test
				void shouldSetEmail() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getEmail()).isEqualTo(AbsenderTestFactory.EMAIL);
				}

				@Test
				void shouldSetTelefon() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getTelefon()).isEqualTo(AbsenderTestFactory.TELEFON);
				}

				@Test
				void shouldSetHyperlink() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getHyperlink()).isEqualTo(AbsenderTestFactory.HYPERLINK);
				}

				@Test
				void shouldSetMandant() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getMandant()).isEqualTo(AbsenderTestFactory.MANDANT);
				}

				@Test
				void shouldSetDienst() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getDienst()).isEqualTo(AbsenderTestFactory.DIENST);
				}

				@Test
				void shouldSetGemeindeschluessel() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getGemeindeschluessel().getSchluessel()).isEqualTo(AbsenderTestFactory.GEMEINDE_SCHLUESSEL);
				}

				@Test
				void shouldSetPostkorbId() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getPostkorbId()).isEqualTo(AbsenderTestFactory.ABSENDER_POSTKORB_ID);
				}

				@Test
				void shouldSetGemeindeschluesselTabelle() {
					var result = mapToNachrichtenKopf().getAbsender();

					assertThat(result.getGemeindeschluessel().getTabelle()).isEqualTo(BayernIdMessageMapper.TABELLE_NUMMER_GEMEINDE_SCHLUESSEL);
				}

			}

			private NachrichtenKopfType mapToNachrichtenKopf() {
				var bspNachricht = mapToBspNachricht();
				assertThat(bspNachricht.getNachrichtenKopf()).isNotNull();
				return bspNachricht.getNachrichtenKopf();
			}
		}

		@Nested
		class TestNachrichtenInhalt {

			@Test
			void shouldSetBetreff() {
				var result = mapToNachrichtenInhalt();

				assertThat(result.getBetreff()).isEqualTo(BayernIdMessageTestFactory.SUBJECT);
			}

			@Test
			void shouldSetStorkQaaLevel() {
				var result = mapToNachrichtenInhalt();

				assertThat(result.getStorkQaaLevel().name()).isEqualTo(BayernIdMessageTestFactory.STORK_QAA_LEVEL);
			}

			@Test
			void shouldSetVorgangsId() {
				var result = mapToNachrichtenInhalt();

				assertThat(result.getZuVorgang().getVorgangsId()).isEqualTo(BayernIdMessageTestFactory.VORGANG_ID);
			}

			@Nested
			class TestMapFreiText {

				@Test
				void shouldSetEncodingSchluessel() {
					var result = mapToFreiText();

					assertThat(result.getEncoding().getSchluessel()).isEqualTo(BayernIdMessageMapper.TEXT_ENCODING_PLAIN_TEXT);
				}

				@Test
				void shouldSetEncodingTabelle() {
					var result = mapToFreiText();

					assertThat(result.getEncoding().getTabelle()).isEqualTo(BayernIdMessageMapper.TABELLE_NUMMER_TEXT_ENCODING);
				}

				@Test
				void shouldSetText() {
					var result = mapToFreiText();

					assertThat(result.getText()).isEqualTo(BayernIdMessageTestFactory.TEXT);
				}

				private FreiTextType mapToFreiText() {
					var bspNachricht = mapToBspNachricht();
					assertThat(bspNachricht.getNachrichtenInhalt().getFreiText()).isNotNull();
					return bspNachricht.getNachrichtenInhalt().getFreiText();
				}
			}

			@Nested
			class TestMapAttachments {

				@Test
				void shouldSetName() {
					var result = mapToDataContainerType();

					assertThat(result.get(0).getFileName()).isEqualTo(AttachmentTestFactory.FILE_NAME);
				}

				@Test
				void shouldSetFileType() {
					var result = mapToDataContainerType();

					assertThat(result.get(0).getFileType().getSchluessel()).isEqualTo(AttachmentTestFactory.TYPE);
				}

				@Test
				void shouldSetFielTypeTable() {
					var result = mapToDataContainerType();

					assertThat(result.get(0).getFileType().getTabelle()).isEqualTo(BayernIdMessageMapper.TABELLE_NUMMER_MIME_TYPE);
				}

				@Test
				void shouldSetInhalt() {
					var result = mapToDataContainerType();

					var attachmentInputStream = result.get(0);
					assertThat(attachmentInputStream.getInhalt()).isEqualTo(AttachmentTestFactory.CONTENT);
				}

				@SneakyThrows
				@Test
				void shouldCloseSourceStream() {
					var inputStream = spy(new ByteArrayInputStream(AttachmentTestFactory.CONTENT));

					var r1 = mapper.toDataContainerType(AttachmentTestFactory.createBuilder().content(inputStream).build());

					verify(inputStream).close();
				}

				private List<DataContainerType> mapToDataContainerType() {
					return mapToNachrichtenInhalt().getDataContainer();
				}
			}

			private NachrichtenInhaltType mapToNachrichtenInhalt() {
				var bspNachricht = mapToBspNachricht();
				assertThat(bspNachricht.getNachrichtenInhalt()).isNotNull();
				return bspNachricht.getNachrichtenInhalt();
			}
		}

		private BspNachricht mapToBspNachricht() {
			return mapper.toBspNachricht(BayernIdMessageTestFactory.create());
		}
	}

	@Nested
	class TestMapToBayernIdResponse {

		@Test
		void shouldSetSuccess() {
			var result = mapToBayernIdResponse();

			assertThat(result.isSuccess()).isTrue();
		}

		@Test
		void shouldSetStatus() {
			var result = mapToBayernIdResponse();

			assertThat(result.getStatus()).isEqualTo(BspQuitungTestFactory.STATUS_SCHLUESSEL);
		}

		@Test
		void shouldSetMessage() {
			var result = mapToBayernIdResponse();

			assertThat(result.getMessage()).isEqualTo(BspQuitungTestFactory.ERGAENZENDE_HINWEISE);
		}

		private BayernIdResponse mapToBayernIdResponse() {
			return mapper.toBayernIdResponse(BspQuitungTestFactory.create());
		}
	}
}